/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

//---PACKAGE DECLARATION---
package uk.org.iscream.cms.server.client.monitors;

//---IMPORTS---
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import uk.org.iscream.cms.server.client.*;
import uk.org.iscream.cms.server.core.*;
import uk.org.iscream.cms.util.*;
import uk.org.iscream.cms.server.componentmanager.*;

/**
 * This Monitor watches heartbeats.
 * It generates an alert when a heartbeat that is expected
 * does not arrive. Unlike all the other monitors, this one
 * is driven by an event *not* occuring, rather than an
 * event occuring. This means it must be actively checking
 * for missing heartbeat's, and thus has an extra inner class
 * thread.
 *
 * This originally took "heartbeat" packets, but they've now
 * been deprecated. Instead we look at UDP packets, or, rather
 * the lack of them :-)
 *
 * @author  $Author: tdb $
 * @version $Id: Heartbeat__Monitor.java,v 1.25 2003/02/24 20:18:48 tdb Exp $
 */
public class Heartbeat__Monitor extends MonitorSkeleton {

//---FINAL ATTRIBUTES---

    /**
     * The current CVS revision of this class
     */
    public final String REVISION = "$Revision: 1.25 $";
    
    /**
     * A description of this monitor
     */
    public final String DESC = "Monitors Heartbeats.";
    
    /**
     * The default (used if not configured) period at
     * which to check for old heartbeats. (in seconds)
     */
    public final int DEFAULT_CHECK_PERIOD = 60;
    
//---STATIC METHODS---

//---CONSTRUCTORS---
    
    /**
     * Constructs a new Heartbeat monitor, and starts off
     * the worker thread.
     */
    public Heartbeat__Monitor() {
        super();
        createInitialHosts();
        new HeartbeatWorker().start();
    }

//---PUBLIC METHODS---
    
    /**
     * Analyse a packet of data. In this case, this will just
     * register the fact that a heartbeat has arrived.
     *
     * @param packet The packet of data to analyse
     */
    public void analysePacket(XMLPacket packet) {
        String source = packet.getParam("packet.attributes.machine_name");
        if (!_hosts.containsKey(source)) {
            synchronized(this) {
                _hosts.put(source, new HeartbeatHolder(new Register(source, _name)));
            }
        }
        HeartbeatHolder lastHeartbeat = (HeartbeatHolder) _hosts.get(source);
        lastHeartbeat.setLastHeartbeat(System.currentTimeMillis()/1000);
    }
    
    /**
     * Overrides the {@link java.lang.Object#toString() Object.toString()}
     * method to provide clean logging (every class should have this).
     *
     * This uses the uk.org.iscream.cms.util.NameFormat class
     * to format the toString()
     *
     * @return the name of this class and its CVS revision
     */
    public String toString() {
        return FormatName.getName(
            _name,
            getClass().getName(),
            REVISION);
    }

    /**
     * return the String representation of what the monitor does
     */
    public String getDescription(){
        return DESC;
    }

//---PRIVATE METHODS---
    
    /**
     * Checks whether the time since the last heartbeat
     * is beyond the threshold(s).
     *
     * @param timeSinceLastHB a long time since the last heartbeat arrived
     * @param reg the Register for this host
     * @return the level which has been breached, if any
     */
    private int checkAttributeThreshold(long timeSinceLastHB, Register reg) {
        for(int thresholdLevel = Alert.thresholdLevels.length - 1; thresholdLevel >= 0; thresholdLevel--) {
            if (reg.getThreshold(thresholdLevel) != -1.0) {
                if (((long) reg.getThreshold(thresholdLevel)) < timeSinceLastHB) {
                    return thresholdLevel;
                }
            }
        }
        return Alert.thresholdNORMAL;
    }
    
    /**
     * Gets an initial list of hosts from the config
     * and adds a fake set of heartbeats for them.
     * If the hosts don't respond within the timeout
     * period an alert will be raised.
     *
     * The effect of this is to allow us to know about
     * hosts which weren't on when we started up, and
     * will thus never have generated a heartbeat - yet
     * will still want to know they're not responding.
     */
    private void createInitialHosts() {
        // get the initial list of hosts from the config
        String initialHosts = "";
        try {
            initialHosts = _cp.getProperty(_name, "Monitor.Heartbeat.initialHosts");
        } catch (PropertyNotFoundException e) {
            // just leave initialHosts empty
            _logger.write(Heartbeat__Monitor.this.toString(), Logger.DEBUG, "No initial list of hosts set, defaulting to none.");
        }
        
        // parse through the initial hosts adding them
        StringTokenizer st = new StringTokenizer(initialHosts, ";");
        while (st.hasMoreTokens()) {
            String source = st.nextToken();
            // check if they already exist, don't want to add them twice
            if (!_hosts.containsKey(source)) {
                synchronized(this) {
                    _hosts.put(source, new HeartbeatHolder(new Register(source, _name)));
                }
            }
            HeartbeatHolder lastHeartbeat = (HeartbeatHolder) _hosts.get(source);
            // set a "fake" heartbeat
            lastHeartbeat.setLastHeartbeat(System.currentTimeMillis()/1000);
        }
    }

//---ACCESSOR/MUTATOR METHODS---
    
    /**
     * Returns a reference to the Queue we're getting data
     * from. This is specific to this monitor.
     *
     * @return a reference to a Queue to get data from
     */
    protected Queue getQueue() {
        return MonitorManager.getInstance().getDataQueue();
    }
    
//---ATTRIBUTES---

    /**
     * This is the friendly identifier of the
     * component this class is running in.
     * eg, a Filter may be called "filter1",
     * If this class does not have an owning
     * component,  a name from the configuration
     * can be placed here.  This name could also
     * be changed to null for utility classes.
     */
    private String _name = "Heartbeat";
    
    /**
     * A reference to the configuration proxy in use
     */
    private ConfigurationProxy _cp = ConfigurationProxy.getInstance();
    
    /**
     * A HashMap of hosts, with associated HeartbeatHolder's.
     */
    private HashMap _hosts = new HashMap();
    
    /**
     * A reference to the system logger.
     */
    private Logger _logger = ReferenceManager.getInstance().getLogger();

//---STATIC ATTRIBUTES---

//---INNER CLASSES---
    
    /**
     * This inner class simply holding some information
     * about a specific host.
     */
    private class HeartbeatHolder {
        
        /**
         * Construct a new HeartbeatHolder.
         */
        public HeartbeatHolder(Register register) {
            _register = register;
        }
        
        /**
         * Set the time of the last heartbeat
         */
        public void setLastHeartbeat(long lastHeartbeat) {
            _lastHeartbeat = lastHeartbeat;
        }
        
        /**
         * Get the time of the last heartbeat
         */
        public long getLastHeartbeat() {
            return _lastHeartbeat;
        }
        
        /**
         * Get the Register
         */
        public Register getRegister() {
            return _register;
        }
        
        /**
         * last heartbeat time
         */
        private long _lastHeartbeat;
        
        /**
         * register ref
         */
        private Register _register;
    }
    
    /**
     * This worker thread just checks all the hosts and then
     * waits a period of time before doing it again. It sends
     * Alerts as required.
     */
    private class HeartbeatWorker extends Thread {
        
        /**
         * The main run method of this worker thread. It simply
         * checks through all the hosts it has stored, running
         * the analyseHB method on each. It then removes any
         * that have passed a FINAL, and waits a (configured)
         * length of time before doing it again.
         */
        public void run() {
            ConfigurationProxy cp = ConfigurationProxy.getInstance();
            while(true) {
                // this cycle period of this monitor's checks
                int checkPeriod = 0;
                try {
                    checkPeriod = Integer.parseInt(cp.getProperty(_name, "Monitor.Heartbeat.checkPeriod"));
                } catch (PropertyNotFoundException e) {
                    checkPeriod = DEFAULT_CHECK_PERIOD;
                    _logger.write(Heartbeat__Monitor.this.toString(), Logger.WARNING, "Monitor.Heartbeat.checkPeriod value unavailable using default of " + checkPeriod + " seconds");
                } catch (NumberFormatException e) {
                    checkPeriod = DEFAULT_CHECK_PERIOD;
                    _logger.write(Heartbeat__Monitor.this.toString(), Logger.WARNING, "Erronous Monitor.Heartbeat.checkPeriod value in configuration using default of " + checkPeriod + " seconds");
                }
                
                synchronized(Heartbeat__Monitor.this) {
                    // perform the checks (use HB hash, although they *should* be the same)
                    Iterator i = _hosts.keySet().iterator();
                    while(i.hasNext()) {
                        // get host
                        String source = (String) i.next();
                        // check it
                        boolean remove = analyseHB(source);
                        // remove it if it's passed a FINAL
                        if(remove) {
                            i.remove();
                        }
                    }
                }
                
                // wait a while
                try {Thread.sleep(checkPeriod * 1000);} catch (InterruptedException e) {}
            }
        }
        
        /**
         * Analyses a given host's state, and if need be generates
         * a relevant Alert. Note that it also checks if the last
         * alert sent is FINAL, in which case it returns true to
         * indicate removal of this host.
         *
         * @param source the host to check
         * @return whether this host can be deleted
         */
        private boolean analyseHB(String source) {
            ConfigurationProxy cp = ConfigurationProxy.getInstance();
            HeartbeatHolder hbHolder = (HeartbeatHolder) _hosts.get(source);
            Register reg = hbHolder.getRegister();
            
            // get host's HB interval (seconds)
            // this should always exist, thus we set to 0
            int hostHBinterval = 0;
            try {
                hostHBinterval = Integer.parseInt(cp.getProperty("Host."+source, "Host.UDPUpdateTime"));
            } catch (PropertyNotFoundException e) {
                hostHBinterval = 0;
                _logger.write(Heartbeat__Monitor.this.toString(), Logger.WARNING, "UDPUpdateTime value unavailable using default of " + hostHBinterval + " seconds");
            } catch (NumberFormatException e) {
                hostHBinterval = 0;
                _logger.write(Heartbeat__Monitor.this.toString(), Logger.WARNING, "Erronous UDPUpdateTime value in configuration using default of " + hostHBinterval + " seconds");
            }
            
            // get host's last HB time (seconds)
            long lastHeartbeat = hbHolder.getLastHeartbeat();
            // time since last heartbeat (seconds)
            long timeSinceLastHB = (System.currentTimeMillis()/1000) - lastHeartbeat;
            // time since (or until if negative) the expected heartbeat
            long timeSinceExpectedHB = timeSinceLastHB - (long) hostHBinterval;
            
            // best do a check in case the expected heartbeat is in the future
            if(timeSinceExpectedHB < 0) {
                timeSinceExpectedHB = 0;
            }
            
            // find out the threshold level we're at
            int newThreshold = checkAttributeThreshold(timeSinceExpectedHB, reg);
            
            // process the alert
            Heartbeat__Monitor.this.processAlert(newThreshold, "Heartbeat", reg, source, String.valueOf(timeSinceExpectedHB));
            
            if(reg.getLastAlertLevel() == Alert.alertFINAL) {
                return true;
            }
            return false;
        }
    }
}
