/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

//---PACKAGE DECLARATION---
package uk.org.iscream.cms.server.clientinterface;

//---IMPORTS---
import uk.org.iscream.cms.util.*;
import uk.org.iscream.cms.server.componentmanager.*;
import uk.org.iscream.cms.server.core.*;
import java.net.Socket;
import java.net.ServerSocket;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.InputStreamReader;


/**
 * Acts as a Control Handler to a TCP based client.
 *
 * @author  $Author: tdb $
 * @version $Id: TCPControlHandler.java,v 1.28 2003/02/05 16:43:46 tdb Exp $
 */
class TCPControlHandler extends Thread {

//---FINAL ATTRIBUTES---

    /**
     * The current CVS revision of this class
     */
    public final String REVISION = "$Revision: 1.28 $";
    
    /**
     * This is our protocol version. It is hardcoded to ensure
     * that it doesn't get accidently changed. A change to the
     * protocol would also require changing this classes code !
     */
    public static final String PROTOVER = "PROTOCOL 1.1";
    
//---STATIC METHODS---

//---CONSTRUCTORS---
    
    /**
     * Construct a new TCPControlHandler, and setup the reader
     * and writer for the new Socket.
     *
     * @param socket The Socket connected to the new Client
     * @param packetSorter A reference to the PacketSorter in the component
     */
    public TCPControlHandler(Socket socket, PacketSorter packetSorter) throws IOException {
        // set the Thread name
        setName("clientinterface.TCPControlHandler");
        
        _socket = socket;
        _packetSorter = packetSorter;
        // setup the reader & writer
        _socketIn = new BufferedReader(new InputStreamReader(_socket.getInputStream()));
        _socketOut = new PrintWriter(_socket.getOutputStream(), true);
        // set the default hostlist to "all host", for backward compatibility
        _hostList = "";
        _logger.write(toString(), Logger.SYSINIT, "created");
    }
    
//---PUBLIC METHODS---
    
    /**
     * This method initiates the thread, setting things up, and then
     * reading commands from the Client. It handles setting up of the
     * DataHandler, and clean shutting down.
     */
    public void run() {
        boolean run = true;
        // Tell the client our protocol, and ask it for it's name
        try {
            send(PROTOVER);
            _clientName = _socketIn.readLine();
            if(_clientName==null) {
                throw new IOException("Fatal error reading from client");
            }
            sendOK();
        }
        catch(IOException e) {
            _logger.write(toString(), Logger.FATAL, "Fatal error, shutdown pending");
            run=false;
        }
        
        _logger.write(toString(), Logger.DEBUG, "Client has connected: "+_clientName);
        
        // loop until we decide to shutdown (run=false)
        while(run) {
            try {
                String cmd = _socketIn.readLine();
                if(cmd==null) {
                    throw new IOException("Fatal error reading from client");
                }
                // make a decision about what to do
                if(cmd.equals("STARTCONFIG")) {
                    // get the configuration for this client
                    ConfigurationProxy cp = ConfigurationProxy.getInstance();
                    sendOK();
                    // get properties
                    cmd = _socketIn.readLine();
                    if(cmd==null) {
                        throw new IOException("Fatal error reading from client");
                    }
                    // provide all the requested properties
                    while(!cmd.equals("ENDCONFIG")) {
                        int midPoint = cmd.indexOf(";");
                        String config, property;
                        if(midPoint != -1) {
                            try {
                                config = cmd.substring(0, midPoint);
                                property = cmd.substring(midPoint+1, cmd.length());
                                if(!config.equals("") && !property.equals("")) {
                                    // client is restricted to these properties and configs
                                    if( (property.startsWith("Client.") || property.startsWith("Host."))
                                      && (config.startsWith("Client.") || config.startsWith("Host.")) ) {
                                        try {
                                            String returnedProperty = cp.getProperty(config, property);    
                                            send(returnedProperty);
                                        }
                                        catch (PropertyNotFoundException e) {
                                            sendERROR();
                                        }
                                    }
                                    else {
                                        sendERROR();
                                    }
                                }
                                else {
                                    sendERROR();
                                }
                            }
                            catch (IndexOutOfBoundsException e) {
                                sendERROR();
                            }
                        }
                        else {
                            sendERROR();
                        }
                        cmd = _socketIn.readLine();
                        if(cmd==null) {
                            throw new IOException("Fatal error reading from client");
                        }
                    }
                    sendOK();
                    _logger.write(toString(), Logger.DEBUG, "Client has been configured");
                    
                }
                else if(cmd.equals("STARTDATA")) {
                    // if we don't have a DataHandler, set one up
                    if(_dataHandler == null) {
                        // This ACL isn't dynamic like the others, but is instead
                        // recreated each time a client connects. Maybe this ACL
                        // could instead be done to allow only the correct client
                        // to connect to the data channel?
                        
                        // get our ACL from the configuration
                        ACL acl = null;
                        try {
                            String stringACL = ConfigurationProxy.getInstance().getProperty("ClientInterface", "ClientInterface.TCPDataChannelACL");
                            acl = new ACL(stringACL);
                        }
                        catch(PropertyNotFoundException e) {
                            _logger.write(toString(), Logger.WARNING, "No ACL found for ClientInterface (data channel listener): " + e);
                        }
                        // create a serversocket
                        // use an ACLServerSocket if we have an ACL
                        ServerSocket ss;
            			if(acl != null) {
            			    ss = new ACLServerSocket(acl, 0);
            			}
            			else {
                            ss = new ServerSocket(0);
                        }
                        // get the port
                        String port = String.valueOf(ss.getLocalPort());
                        // tell the client the port
                        send(port);
                        // wait for the client to connect
                        Socket s = ss.accept();
                        // when we get the Socket back, give it to the data thread
                        TCPDataHandler dh = new TCPDataHandler(s);
                        // register the DataHandler's queue, giving the host list
                        _packetSorter.register(dh.getQueue(), _hostList);
                        try {
                            // startup a monitor on the DataHandler's queue
                            ConfigurationProxy cp = ConfigurationProxy.getInstance();
                            int queueMonitorInterval = Integer.parseInt(cp.getProperty("ClientInterface", "Queue.MonitorInterval"));
                            String queueName = _name + " TCPHandler:"+_socket.getInetAddress().getHostName();
                            dh.getQueue().startMonitor(queueMonitorInterval*1000, _packetSorter.getQueue(), queueName);
                        } catch (PropertyNotFoundException e) {
                            _logger.write(toString(), Logger.WARNING, "failed to find queue monitor config, disabling. " + e);
                        }
                        // start up the DataHandler
                        dh.start();
                        // Hold a reference to the DataHandler, so we can stop it later
                        _dataHandler = dh;
                        sendOK();
                        _logger.write(toString(), Logger.DEBUG, "Data stream started at Clients Request on port: "+port);
                    }
                    else {
                        sendERROR();
                    }
                }
                else if(cmd.equals("STOPDATA")) {
                    // attempt to close the data channel
                    if(closeData()) {
                        sendOK();
                        _logger.write(toString(), Logger.DEBUG, "Data stream stopped at Clients Request");
                    }
                    else {
                        sendERROR();
                    }
                }
                else if(cmd.equals("SETHOSTLIST")) {
                    // we can only set the host list when
                    // the DataHandler is not connected
                    if(_dataHandler == null) {
                        sendOK();
                        // read and set the host list
                        cmd = _socketIn.readLine();
                        if(cmd==null) {
                            throw new IOException("Fatal error reading from client");
                        }
                        _hostList = cmd;
                        sendOK();
                    }
                    else {
                        sendERROR();
                    }
                }
                else if(cmd.equals("DISCONNECT")) {
                    // we going to disconnect, so lets stop the main loop
                    run=false;
                    // if there is a DataHandler, we'd best shut it down
                    if(closeData()) {
                        _logger.write(toString(), Logger.DEBUG, "Data stream stopped at Clients Request");
                    }
                    // say bye to the client
                    sendOK();
                    // close the reader, writer and Socket
                    _socketIn.close();
                    _socketOut.close();
                    _socket.close();
                    _logger.write(toString(), Logger.DEBUG, "Closing at Clients Request");
                }
                else {
                    sendERROR();
                }
            }
            catch(IOException e) {
                // if we get an exception, the client has probably left, so we stop
                run=false;
                _logger.write(toString(), Logger.FATAL, "Fatal communication error, shutdown pending");
            }
        }
        // we'll close any DataHandlers here that shouldn't be open still
        if(closeData()) {
            _logger.write(toString(), Logger.DEBUG, "Data stream stopped due to fatal client error");
        }
        _logger.write(toString(), Logger.DEBUG, "Shutting down Control Handler, client has gone.");
    }
  
    /**
     * Overrides the {@link java.lang.Object#toString() Object.toString()}
     * method to provide clean logging (every class should have this).
     *
     * This uses the uk.org.iscream.cms.util.NameFormat class
     * to format the toString()
     *
     * @return the name of this class and its CVS revision
     */
    public String toString() {
        return FormatName.getName(
            _name,
            getClass().getName(),
            REVISION);
    }

//---PRIVATE METHODS---

    /**
     * Attempt to close down the DataHandler. This will return
     * true if it managed to close it down, or false if there
     * wasn't a DataHandler to close.
     *
     * @return whether the channel could be closed
     */
    private boolean closeData() {
        // if we have a DataHandler, shut it down
        if(_dataHandler != null) {
            // Deregister the DataHandler, giving the host list
            _packetSorter.deregister(_dataHandler.getQueue(), _hostList);
            // Shut down the data handler
            _dataHandler.shutdown();
            // Destroy our reference, leaving it for the GC
            _dataHandler = null;
            return true;
        }
        else {
            return false;
        }
    }
    
    /**
     * Send an OK message to the outgoing Socket
     */
    private void sendOK() {
        send("OK");
    }
    
    /**
     * Send an ERROR message to the outgoing Socket.
     */
    private void sendERROR() {
        send("ERROR");
    }
    
    /**
     * Send any String message to the outgoing Socket.
     *
     * @param message The message/command to send.
     */
    private void send(String message) {
        _socketOut.println(message);
    }

//---ACCESSOR/MUTATOR METHODS---

//---ATTRIBUTES---

    /**
     * This is the friendly identifier of the
     * component this class is running in.
     * eg, a Filter may be called "filter1",
     * If this class does not have an owning
     * component,  a name from the configuration
     * can be placed here.  This name could also
     * be changed to null for utility classes.
     */
    private String _name = ClientInterfaceMain.NAME;
    
    /**
     * This holds a reference to the
     * system logger that is being used.
     */
    private Logger _logger = ReferenceManager.getInstance().getLogger();
    
    /**
     * A reference to the Configuration Manager the system is using
     */
    private ConfigurationManager _configManager = ReferenceManager.getInstance().getCM();
    
    /**
     * The socket we are talking on
     */
    private Socket _socket;

    /**
     * A hook to the inbound data from the socket
     */
    private BufferedReader _socketIn;

    /**
     * A hook to the outbound stream for the socket
     */
    private PrintWriter _socketOut;
    
    /**
	 * A reference to the PacketSorter.
	 */
    private PacketSorter _packetSorter;
    
    /**
     * A reference to the DataHandler, if there is one
     */
    private TCPDataHandler _dataHandler;
    
    /**
     * The name of the Client we're connected to
     */
    private String _clientName;
    
    /**
     * The host list the Client has requested
     */
    private String _hostList;
    
//---STATIC ATTRIBUTES---

}
