/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

//---PACKAGE DECLARATION---
package uk.org.iscream.cms.server.componentmanager;

//---IMPORTS---
import java.util.*;
import java.io.*;
import uk.org.iscream.cms.util.*;

/**
 * The component manager is the starting point for all
 * server side components of the iscream system.
 * It loads its initial system configuration from the
 * default properties file, it then starts all the iscream
 * components as specified in the default.properties under
 * uk.org.iscream.cms.server.ComponentList
 *
 * @author  $Author: tdb $
 * @version $Id: ComponentManager.java,v 1.42 2003/02/05 16:43:46 tdb Exp $
 */
public class ComponentManager {

//---FINAL ATTRIBUTES---

    /**
     * The current CVS revision of this class
     */
    public static final String REVISION = "$Revision: 1.42 $";
    
    /**
     * The toString() of this class
     * As it won't be instatiated, this is needed.
     * Not also that we pass a null as the class name (as we are static)
     */
    public static final String toString = FormatName.getName("ComponentManager", null, REVISION);
    
    /**
     * The default location of the properties file for the system
     */
    public static final String DEFAULTPROPERTIES = "./etc/default.properties";
    
    /**
     * The default time to wait before retrying 
     * component.
     */
    public static final int DEFAULT_COMPONENT_START_TIMEOUT = 5;
    
//---STATIC METHODS---

    /**
     * The main method which starts the components as
     * listed in the default.properties file.
     *
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        System.out.println("-----------------------------------------");
        System.out.println("--- i-scream Server Component Manager ---");
        System.out.println("---   (c) 2001 The i-scream Project   ---");
        System.out.println("---    (http://www.i-scream.org.uk)   ---");
        System.out.println("-----------------------------------------");
        System.out.println("---          Starting System          ---");        
        
        // get the command line args
        String defaultProperties = DEFAULTPROPERTIES;
        String filterName = null;
        String filterManagerName = null;
        for(int i=0; i < args.length; i++) {
            if(args[i].equals("-h")) {
                usage();
            }
            else if(args[i].equals("-f")) {
                if(++i < args.length) {
                    filterName = args[i];
                }
                else {
                    usage();
                }
            }
            else if(args[i].equals("-fm")) {
                if(++i < args.length) {
                    filterManagerName = args[i];
                }
                else {
                    usage();
                }
            }
            else if(args[i].equals("-l")) {
                if(++i < args.length) {
                    defaultProperties = args[i];
                }
                else {
                    usage();
                }
            }
            else {
                usage();
            }
        }

        // load the default properties file into the system properties
        System.out.println(toString + ": initialising - using " + defaultProperties);
        try {
            Properties initProperties = new Properties(System.getProperties());
            initProperties.load(new FileInputStream(new File(defaultProperties)));
            System.setProperties(initProperties);
        } catch (Exception e) {
            System.err.println(toString + ": ERROR " + e.getMessage());
            usage();
        }

        // continue to bring the system up
        System.out.println(toString + ": coming up");
        
        // start the ORB by initialising the ReferenceManager
        ReferenceManager refman = ReferenceManager.getInstance();
        
        // now the ORB is running, we need to activate our RootPOA
        // so that we can start serving requests once servants start up
        refman.activatePOA();
        
        // get the list of components
        String componentList = System.getProperty("uk.org.iscream.cms.server.ComponentList");
        StringTokenizer st = new StringTokenizer(componentList, ";");
        _componentsToStart = new LinkedList();
        
        // this could be done using reflection
        // but..well..we don't ;-p
        while (st.hasMoreTokens()){
            String componentName = st.nextToken();
            Component component = null;
            
            
            // ### This is where the list of supported components is checked! ###
            if (componentName.equalsIgnoreCase("core")) {
                component = new uk.org.iscream.cms.server.core.Core();
            // note the passing of the FilterManagers's name in its constructor
            } else if (componentName.equalsIgnoreCase("filtermanager")) {
                component = new uk.org.iscream.cms.server.filtermanager.FilterManager(filterManagerName);
            } else if (componentName.equalsIgnoreCase("rootfilter")) {
                component = new uk.org.iscream.cms.server.rootfilter.RootFilter();
            } else if (componentName.equalsIgnoreCase("dbinterface")) {
                component = new uk.org.iscream.cms.server.dbinterface.DBInterface();
            } else if (componentName.equalsIgnoreCase("clientinterface")) {
                component = new uk.org.iscream.cms.server.clientinterface.ClientInterfaceMain();
            // note the passing of the Filter's name in its constructor
            } else if (componentName.equalsIgnoreCase("filter")) {
                component = new uk.org.iscream.cms.server.filter.FilterMain(filterName);
            } else if (componentName.equalsIgnoreCase("client")) {
                component = new uk.org.iscream.cms.server.client.ClientMain();
            }
            // ###  Add new component constructors in the above section! ###
            
            // build the list of components to start
            if (component != null) {
                _componentsToStart.add(component);    
            } else {
                System.err.println(toString + ": WARNING unsupported component not started: "+componentName);
            }
        }
        
        

        // get the value for timeout
        
        String confTimeout = null;
        try {
            confTimeout = System.getProperty("uk.org.iscream.cms.server.ComponentTimeout");
            confTimeout.trim();
            _startTimeout = Integer.parseInt(confTimeout);
        } catch (NumberFormatException e) {
            _startTimeout = DEFAULT_COMPONENT_START_TIMEOUT;
            System.err.println(toString + ": unable to read uk.org.iscream.cms.server.ComponentTimeout value (" + confTimeout + "), using default!");
        }
        System.out.println(toString + ": using component start timeout of " + _startTimeout + " seconds");
        
        // startup the system components
        startUp();

	    // block on the ORB...in time, management functionality can be placed here.
	    // if we detect a CORBA communication error, we'll restart all the components.
	    // !! this doesn't appear to work !!
	    while(true) {
    	    try {
                refman.getORB().run();
            } catch (org.omg.CORBA.COMM_FAILURE e) {
                System.out.println(toString + ": WARNING CORBA communications failure - " + e.getMessage());
                System.out.println(toString + ": WARNING CORBA connections have been lost - attempting to restart!");
            }
            startUp();
        }
    }

    /**
     * Starts the components as obtained from the configuration.
     * This method calls the start() methods on all the components.
     * If a component fails to start due to a CORBA communication
     * problem, then it catches this and tries to start it again
     * according to uk.org.iscream.cms.server.ComponentTimeout time.
     *
     * If the server dies and CORBA connections are lost, this method
     * is called again.
     */
    private static void startUp() {
        // now we try and start the beast up        
        boolean tryAgain = true;
        Component component = null;
        
        // keep trying until we've started all the components.
        // maybe add support for a limited number of retries
        while(tryAgain) {
            Iterator i = _componentsToStart.iterator();
            LinkedList failedComponents = new LinkedList();
            // go through all the components
            while(i.hasNext()) {
    	        // get a refence to the component
    	        component = (Component) i.next();
    	        System.out.println(toString + ": dependency checking component - " + component.toString());
    	        
    	        // check it's dependencies
    	        boolean depOK = component.depCheck();
    	        if(depOK) {
    	            System.out.println(toString + ": starting component - " + component.toString());
    	            // it should be ok to start the component
        	        try {    
        	            // start the component
                        component.start();
                    } catch (ComponentStartException e) {
                        // if we get this then there was a problem
                        // that we can't recover from
                        System.err.println(toString + ": ERROR starting component - " + component.toString());
                        System.err.println(toString + ": component reports - " + e.getMessage());
                        System.exit(1);
                    }
                }
                else {
                    // it seems the depedencies failed
                    // so we want to try again after a delay
                    System.err.println(toString + ": WARNING Component reported dependency failure");
                    System.err.println(toString + ": This could be because it can't communicate with components it needs.");
                    
                    // make a list of the failed components
                    failedComponents.add(component);
                }
            }
            
            // if we had some failed components that we can retry
            if (failedComponents.size() > 0) {
                System.err.println(toString + ": WARNING One or more components failed to start correctly.");
                System.err.println(toString + ": Will try again in " + _startTimeout + " seconds");
            
                // our list is now the failed list
                _componentsToStart = failedComponents;
                
                // sleep for a given timeout
                try {    
                    Thread.sleep(_startTimeout * 1000);
                } catch (InterruptedException e) {
                    // we're not bothered
                }
            // otherwise we started everything
            } else {
                // so we set to exit the loop
                tryAgain = false;
            }
        }
	        
	    System.out.println(toString + ": running");
	}

    /**
     * A simple method to print the usage of this class.
     * It never returns, but instead exits to the system
     * with a value 1, to indicate the system did not start
     * properly.
     */
    public static void usage() {
        System.out.println("USAGE: java uk.org.iscream.cms.server.componentmanager.ComponentManager <option>");
        System.out.println("   or: java -jar iscream.jar <option>");
        System.out.println("WHERE <option>:");
        System.out.println("      -l <filename> - the location of initial system properties");
        System.out.println("                      the default is ./etc/default.properties");
        System.out.println("      -f <name>     - the name of the filter (if there is one configured");
        System.out.println("      -fm <name>    - the name of the filter manager (if there is one configured");
        System.out.println("      -h            - this help screen");
        System.exit(1);
    }

//---CONSTRUCTORS---

//---PUBLIC METHODS---

//---PRIVATE METHODS---

//---ACCESSOR/MUTATOR METHODS---

//---ATTRIBUTES---
    
//---STATIC ATTRIBUTES---

    private static LinkedList _componentsToStart;
    private static int _startTimeout = 0;

}
