/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

//---PACKAGE DECLARATION---
package uk.org.iscream.cms.server.componentmanager;

//---IMPORTS---
import org.omg.CORBA.ORB;
import org.omg.CosNaming.*;
import org.omg.PortableServer.*;
import uk.org.iscream.cms.server.core.*;
import uk.org.iscream.cms.util.*;

/**
 * This class returns references for global system objects.
 * This class is used to create and return references to objects
 * that are required throughout the system.  Most of which
 * are CORBA based.  It also manages the ORB for the component.
 *
 * It is a singleton object with a static method to obtain a
 * reference to it.  This removes the need for passing
 * references to all the objects it contains throughout 
 * a component.
 *
 * @author  $Author: tdb $
 * @version $Id: ReferenceManager.java,v 1.19 2003/02/05 16:43:46 tdb Exp $
 */
public class ReferenceManager {

//---FINAL ATTRIBUTES---
   
    /**
     * The current CVS revision of this class
     */
    public final String REVISION = "$Revision: 1.19 $";
    
//---STATIC METHODS---

    /**
     * This returns a reference to the single instance of the
     * ReferenceManager.
     *
     * This creates the single instance of the ReferenceManager
     * if it does not exist by calling the private constructor.
     *
     * @return a reference to the ReferenceManager
     */
    public synchronized static ReferenceManager getInstance() {
        if (_instance == null) {
            _instance = new ReferenceManager();
        }
        return _instance;
    }

//---CONSTRUCTORS---

    /**
     * This is a private constructor
     * This ensures that the system performs according to a Singleton design pattern
     */
    private ReferenceManager() {
        _orb = ORB.init(new String[] {}, null);
    }

//---PUBLIC METHODS---

    /**
     * Obtains a CORBA reference through the naming service
     * for the named object.
     * Calls the dieWithError() if any exceptions happen!
     *
     * @param name the name of the CORBA object to resolve
     *
     * @return a reference to the object
     */
    public org.omg.CORBA.Object getCORBARef(String name) {
        org.omg.CORBA.Object objRef = null;
        try {
            objRef = getNS().resolve(getNS().to_name(name));
        } catch (org.omg.CosNaming.NamingContextPackage.CannotProceed e) {
            dieWithError("\nCRITICAL:Naming Service Cannot Proceed - when resolving reference to " + name + "!\n" + 
                              "         Please check with your CORBA naming service provider.");
        } catch (org.omg.CosNaming.NamingContextPackage.InvalidName e) {
            dieWithError("\nCRITICAL:Invalid Name - when resolving reference to " + name + "!\n" + 
                              "         Please check with your CORBA naming service provider.");
        } catch (org.omg.CosNaming.NamingContextPackage.NotFound e) {
            recoverWithError("\nCRITICAL:Not Found - when resolving reference to " + name + "!\n" + 
                              "         Please check that this component is running.");
        }
        return objRef;
    }

    /** 
     * Binds a given servant with the given name
     * to the naming service.
     * Calls the dieWithError() if any exceptions happen!
     *
     * @param objRef a reverence to the servant object
     * @param name the name to bind to the naming service with
     */
    public void bindToOrb(org.omg.PortableServer.Servant objRef, String name) {
        try {
            getNS().bind(getNS().to_name(name), getRootPOA().servant_to_reference(objRef));
        } catch (org.omg.PortableServer.POAPackage.WrongPolicy e) {
            dieWithError("\nCRITICAL:Wrong POA Policy - when binding " + name + "!\n" + 
                              "         This indicates an error with your ORB.");
        } catch (org.omg.PortableServer.POAPackage.ServantNotActive e) {
            dieWithError("\nCRITICAL:ServantNotActive - when binding " + name + "!\n" + 
                              "         This indicates an error with your ORB.");
        } catch (org.omg.CosNaming.NamingContextPackage.InvalidName e) {
            dieWithError("\nCRITICAL:Invalid Name - when binding " + name + "!\n" + 
                              "         Please check with your CORBA naming service provider.");
        } catch (org.omg.CosNaming.NamingContextPackage.CannotProceed e) {
            dieWithError("\nCRITICAL:Naming Service Cannot Proceed - when binding " + name + "!\n" + 
                              "         Please check with your CORBA naming service provider.");
        } catch (org.omg.CosNaming.NamingContextPackage.NotFound e) {
            recoverWithError("\nCRITICAL:Naming Service Not Found - when binding " + name + "!\n" + 
                              "         Please check with your CORBA naming service provider.");
        } catch (org.omg.CosNaming.NamingContextPackage.AlreadyBound e) {
            dieWithError("\nCRITICAL:Already Bound - when binding " + name + "!\n" + 
                              "         Another component with this name is already running.");
        }
    }
    
    /**
     * Activates the POA
     * Calls the dieWithError() if any exceptions happen!
     */
    public void activatePOA() {
        try {
            getRootPOA().the_POAManager().activate();
        } catch (org.omg.PortableServer.POAManagerPackage.AdapterInactive e) {
            dieWithError("\nCRITICAL:POA Set Inactive - when trying to activate POA!\n" + 
                              "         This indicates an error with your ORB.");
        }
    }
    
    /**
     * Overrides the {@link java.lang.Object#toString() Object.toString()}
     * method to provide clean logging (every class should have this).
     *
     * This uses the uk.org.iscream.cms.util.NameFormat class
     * to format the toString()
     *
     * @return the name of this class and its CVS revision
     */
    public String toString() {
        return FormatName.getName(
            _name,
            getClass().getName(),
            REVISION);
    }

//---PRIVATE METHODS---

    /**
     * If there are any CORBA errors this method is called with the 
     * error message.
     *
     * The exception this throws is generic to CORBA communication
     * problems, this is a Runtime exception and is typically only
     * explicitly caught in the ComponentManager for the current VM
     * The ComponentManager can then decide what course of action to 
     * take.
     *
     * Also, if it can, it will log the message with the logger.
     *
     * @param message the error message to die with
     */
    private void recoverWithError(String message) throws ComponentCORBAException {
        if (_logger != null) {
            _logger.write(toString(), Logger.WARNING, "component CORBA error - " + message);
        }
        throw new ComponentCORBAException(message);
    }


    /**
     * If there are any CORBA errors this method is called with the 
     * error message.
     *
     * Currently this prints the error, on the local err stream.  Will
     * print to the logger if one is available.
     *
     * ANY CALLS TO THIS METHOD CAUSE
     * THE SYSTEM TO EXIT WITH A NON
     * ZERO CODE!
     *
     * @param message the error message to die with
     */
    private void dieWithError(String message) {
        System.err.println(message);
        if (_logger != null) {
            _logger.write(toString(), Logger.FATAL, message);
        }
        System.exit(1);
    }



//---ACCESSOR/MUTATOR METHODS---

    /**
     * Returns a reference to the ORB
     *
     * @return the reference this class holds
     */
    public ORB getORB() {
        return _orb;
    }

    /**
     * Returns a reference to the Root POA
     * Calls the dieWithError() if any exceptions happen!
     *
     * @return the reference this class holds
     */
    public POA getRootPOA() {
        if (_rootPOA == null) {
            org.omg.CORBA.Object objRef = null;
            try {
                objRef = getORB().resolve_initial_references("RootPOA");
            } catch (org.omg.CORBA.ORBPackage.InvalidName e) {
                dieWithError("\nCRITICAL:Unable to resolve reference to the RootPOA!\n" + 
                                  "         This indicates an error with your ORB.");
            }
            _rootPOA = POAHelper.narrow(objRef);
        }
        return _rootPOA;
    }
    
    /**
     * Returns a reference to the Naming Service
     * Calls the dieWithError() if any exceptions happen!
     *
     * @return the reference this class holds
     */
    public NamingContextExt getNS() {
        if (_ns == null) {
            org.omg.CORBA.Object objRef = null;
            try {
                objRef = getORB().resolve_initial_references("NameService");
            } catch (org.omg.CORBA.ORBPackage.InvalidName e) {
                dieWithError("\nCRITICAL:Unable to resolve reference to the Naming Service!\n" +
                                  "         Please check with your CORBA naming service provider.");
            }
            _ns = NamingContextExtHelper.narrow(objRef);
        }
        // check we managed to talk to the naming service
        if (_ns == null) {
            dieWithError("\nCRITICAL:Unable to resolve reference to the Naming Service!\n" +
                              "         Please check with your CORBA naming service provider.");
        }
        return _ns;
    }
    
    /**
     * Returns a reference to the configuration manager
     * Calls the dieWithError() if any exceptions happen!
     *
     * @return the reference this class holds
     */
    public ConfigurationManager getCM() {
        if (_cm == null) {
            _cm = ConfigurationManagerHelper.narrow(getCORBARef("iscream.ConfigurationManager"));
        }
        // check we managed to talk to the configuration manager
        if (_cm == null) {
            dieWithError("\nCRITICAL:Unable to resolve reference to the Configuration Manager!\n" +
                              "         Please check with your CORBA naming service provider.");
        }
        return _cm;
    }

    /**
     * Returns a reference to the logger
     * This will throw a Error if there is an error
     *
     * @return the reference this class holds
     */
    public Logger getLogger() {
        if (_logger == null) {
            _logger = LoggerHelper.narrow(getCORBARef("iscream.Logger"));
        }
        return _logger;
    }
    
//---ATTRIBUTES---
    
    /**
     * The ORB
     */
    private ORB _orb;
    
    /**
     * The Naming Service
     */
    private NamingContextExt _ns;
    
    /**
     * The Root POA
     */
    private POA _rootPOA;

    /**
     * The configuration manager
     */
    private ConfigurationManager _cm;

    /**
     * The logger
     */
    private Logger _logger;
    
    /**
     * This is the friendly identifier of the
     * component this class is running in.
     * eg, a Filter may be called "filter1",
     * If this class does not have an owning
     * component,  a name from the configuration
     * can be placed here.  This name could also
     * be changed to null for utility classes.
     */
    private String _name = null;

//---STATIC ATTRIBUTES---
    
    /**
     * A reference to the single instance of this class
     */
    private static ReferenceManager _instance = null; 
}
