/* 
 * i-scream central monitoring system
 * http://www.i-scream.org.uk
 * Copyright (C) 2000-2002 i-scream
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

//---PACKAGE DECLARATION---
package uk.org.iscream.cms.server.filter;

//---IMPORTS---
import java.io.*;
import java.util.*;

import uk.org.iscream.cms.util.*;
import uk.org.iscream.cms.server.componentmanager.*;
import uk.org.iscream.cms.server.core.*;

/**
 * This class setups up and manages Plugins in a Filter. A list
 * of plugins to use is specified in the configuration, and these
 * are all loaded upon first use of this class.
 * This is a singleton class.
 *
 * @author  $Author: tdb $
 * @version $Id: PluginFilterManager.java,v 1.13 2003/02/05 16:43:47 tdb Exp $
 */
class PluginFilterManager {

//---FINAL ATTRIBUTES---

    /**
     * The current CVS revision of this class
     */
    public final String REVISION = "$Revision: 1.13 $";
    
//---STATIC METHODS---

    /**
     * Return a reference to the single class.
     * Construct it if it does not already exist, otherwise just return the reference.
     */
    public static PluginFilterManager getInstance() {
        if (_instance == null){
            _instance = new PluginFilterManager();
        }
        return _instance;
    }

//---CONSTRUCTORS---

    /**
     * Private Constructor - this part creates the filter pipeline
     * This is a singleton class, btw.
     */
    private PluginFilterManager(){
        
        _logger.write(toString(), Logger.SYSINIT, "Initialising");
        _logger.write(toString(), Logger.SYSMSG, "Creating filter pipeline for plugin filters ...");
        
        String pluginsPackage, pluginsList;
        
        try {
            // get the configuration for this plug-in setup
            ConfigurationProxy cp = ConfigurationProxy.getInstance();
            pluginsPackage = cp.getProperty("Filter." + FilterMain.NAME, "Filter.PluginsPackage");
            pluginsList = cp.getProperty("Filter." + FilterMain.NAME, "Filter.Plugins");
        } catch (PropertyNotFoundException e) {
            _logger.write(toString(), Logger.WARNING, "Unable to locate property: "+e);
            // setting this to "" will avoid building a plugin list,
            // which is what we want when a property doesn't exist
            pluginsList = "";
            // and set this to "", because the compiler is a git
            pluginsPackage = "";
        }
        
        StringTokenizer st = new StringTokenizer(pluginsList, ";");
        
        while(st.hasMoreTokens()) {
            String className = pluginsPackage + "." + st.nextToken() + _suffix;
            _logger.write(toString(), Logger.DEBUG, "Attempting to create plugin: "+className);
            
            // Create an instance of the specified PluginFilter to include
            // within the filterPipe.  Add it to the filterPipeline
            try {
                PluginFilter pf = (PluginFilter)ClassLoader.getSystemClassLoader().loadClass(className).newInstance();
                _filterPipeline.add(pf);
                _logger.write(toString(), Logger.DEBUG, "Added filter: "+className+" ("+pf.getDescription()+")");
            }
            catch (InstantiationException e){
                _logger.write(toString(), Logger.ERROR, "Failed to instantiate "+className+" to the plugin filter pipeline.");
                _logger.write(toString(), Logger.ERROR, e.getMessage());
            }
            catch (Exception e){
                _logger.write(toString(), Logger.ERROR, "Failed to add "+className+" to the plugin filter pipeline.");
                _logger.write(toString(), Logger.ERROR, e.toString());
            }
        }
        _logger.write(toString(), Logger.SYSMSG, "The filter pipeline has been set up with "+_filterPipeline.size()+" plugin filters.");
        
    }

//---PUBLIC METHODS---

    /**
     * apply all of the filters in the pipeline to the packet.
     * return true if they all accept the packet.
     * return false if any single filter rejects the packet.
     * return true if there are no filters inthe pipeline.
     *
     * @param packet an XMLPacket to be tested
     * @return whether the packet can be passed on
     */
    public boolean runFilters(XMLPacket packet){
        
        // for each filter in the pipeline...
        ListIterator pluginFilters = _filterPipeline.listIterator(0);
        while (pluginFilters.hasNext()){
            PluginFilter filter = (PluginFilter)pluginFilters.next();
            if (!filter.runFilter(packet)){
                return false;
            }
        }
        
        return true;
    }

    /**
     * Overrides the {@link java.lang.Object#toString() Object.toString()}
     * method to provide clean logging (every class should have this).
     *
     * This uses the uk.org.iscream.cms.util.NameFormat class
     * to format the toString()
     *
     * @return the name of this class and its CVS revision
     */
    public String toString() {
        return FormatName.getName(
            _name,
            getClass().getName(),
            REVISION);
    }

//---PRIVATE METHODS---

//---ACCESSOR/MUTATOR METHODS---

//---ATTRIBUTES---
   
    /**
     * file name suffix for plugin filter classes:
     */
    private final String _suffix = "__Plugin";
    
    /**
     * LinkedList for holding the PluginFilter objects (the pipeline).
     */
    private LinkedList _filterPipeline = new LinkedList();
    
    /**
     * A reference to the single instance of this class
     */
    private static PluginFilterManager _instance;

    /**
     * This is the friendly identifier of the
     * component this class is running in.
     * eg, a Filter may be called "filter1",
     * If this class does not have an owning
     * component,  a name from the configuration
     * can be placed here.  This name could also
     * be changed to null for utility classes.
     */
    private String _name = FilterMain.NAME;

    /**
     * This holds a reference to the
     * system logger that is being used.
     */
    private Logger _logger = ReferenceManager.getInstance().getLogger();
    
    /**
     * A reference to the reference manager in use
     */
    private ReferenceManager _refman = ReferenceManager.getInstance();

//---STATIC ATTRIBUTES---

}
